/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package mindex.distance;

import messif.algorithms.AlgorithmMethodException;
import messif.objects.LocalAbstractObject;
import messif.operations.RankingSingleQueryOperation;
import mindex.MetricIndex;

/**
 * This query-PPP distance calculator uses the query-pivot distances as in {@link QueryPivotDistDiffCalculator}
 *  and adds a "correction" to the calculated distance. The correction results in a value that is still 
 *  a lower bound on the distances between the query and objects on the last level (l) but this
 *  lower bound is tighter (better). 
 * @author xnovak8
 */
public class WeightedSumDistCalculatorCorrected extends WeightedSumDistCalculator implements PartialQueryPPPDistanceCalculator {
    
    private float [] corrections;

    public WeightedSumDistCalculatorCorrected(RankingSingleQueryOperation operation, MetricIndex mIndex) throws AlgorithmMethodException {
        super(operation, mIndex);
        initCorrections();
    }
    
    public WeightedSumDistCalculatorCorrected(LocalAbstractObject queryObject, MetricIndex mIndex, float coeficient) throws AlgorithmMethodException {
        super(queryObject, mIndex, coeficient);
        initCorrections();
    }
    
    protected WeightedSumDistCalculatorCorrected(short maxLevel, float [] queryPivotDistances, short [] queryPPP) {
        super(maxLevel, queryPivotDistances, queryPPP);
        initCorrections();
    }
    
    private void initCorrections() {
        corrections = new float[maxLevel + 1];
        // calculate the corrections for individual levels (lengths of PPP)
        for (int lstripe = maxLevel - 1; lstripe >= 0; lstripe--) {
            for (int j = lstripe; j <= maxLevel - 1; j++) {
                corrections[lstripe] += COEFS[j] * queryPivotDistances[queryPP[j - lstripe]];
            }
        }
    }

    // **************   Distance calculations    ************************** //
    
    @Override
    public final float getQueryDistance(final short[] ppp) {
        return getDistance(ppp, 1, getMaxLevel(ppp.length));
    }

    @Override
    public float getPartialQueryDistance(final short[] ppp, final int fromLevel) {
        if (ppp.length == 0) {
            return 0;
        }
        return getDistance(ppp, fromLevel, getMaxLevel(fromLevel - 1 + ppp.length)) - corrections[fromLevel - 1];
    }

    private float getDistance(final short[] ppp, final int fromLevel, final int toLevel) {
        float sum = 0f;
        for (int i = 0; i <= toLevel - fromLevel; i++) {
            sum += (queryPivotDistances[ppp[i]] * COEFS[i + fromLevel - 1]);
        }
        return sum + corrections[toLevel];
    }

    @Override
    public float getPartialQueryDistance(short ppIndex, int fromLevel) {
        return (queryPivotDistances[ppIndex] * COEFS[fromLevel - 1]) - corrections[fromLevel - 1] + corrections[fromLevel];
    }
        
}
