package mindex.distance;

import messif.algorithms.AlgorithmMethodException;
import messif.operations.RankingSingleQueryOperation;
import mindex.MetricIndex;
import mindex.navigation.NextLevelPivotChecker;

/**
 * This query-PPP distance calculator uses the query-pivot distances as in {@link WeightedSumDistCalculator}
 *  and tries to estimate the minimum distance to objects in the tested subtree as possible. It 
 *  actually uses the {@link NextLevelPivotChecker} to look "forward" to pivot indexes actually stored on the next level.
 * 
 * @author xnovak8
 */
public class WeightedSumDistCalculatorTight extends WeightedSumDistCalculator {

    public WeightedSumDistCalculatorTight(RankingSingleQueryOperation operation, MetricIndex mIndex) throws AlgorithmMethodException {
        super(operation, mIndex);
    }
    
    protected WeightedSumDistCalculatorTight(short maxLevel, float [] queryPivotDistances, short [] queryPPP) {
        super(maxLevel, queryPivotDistances, queryPPP);
    }

    // **************   Distance calculations    ************************** //

    @Override
    public final float getQueryDistance(final short[] ppp) {
        return getQueryDistance(ppp, null);
    }

    @Override
    public float getQueryDistance(short[] ppp, NextLevelPivotChecker cell) {
        int length = getMaxLevel(ppp.length);
        float sum = 0f;
        for (int i = 0; i < length; i++) {
            sum += (queryPivotDistances[ppp[i]] * COEFS[i]);
        }
        
        // correction to estimate the minimum possible distance of data with this ppp prefix
        int smallestNonusedIndex = 0;
        int usedFirstIndex = -1;
        for (int i = length; i < maxLevel; i++) {
            if (smallestNonusedIndex == usedFirstIndex) {
                smallestNonusedIndex++;
            }

            int j = 0;
            while (j < length) {
                if (queryPP[smallestNonusedIndex] == ppp[j++]) {
                    if (++smallestNonusedIndex == usedFirstIndex) {
                        smallestNonusedIndex++;
                    }
                    j = 0;
                }
            }

            // check the really existing nodes on the next level
            if (i == length && cell != null) {
                usedFirstIndex = smallestNonusedIndex;
                while (! cell.containsGivenPivot(queryPP[usedFirstIndex])) {
                    usedFirstIndex++;
                }
                sum += queryPivotDistances[queryPP[usedFirstIndex]] * COEFS[i];
            } else {
                sum += queryPivotDistances[queryPP[smallestNonusedIndex++]] * COEFS[i];
            }
        }
        return sum;
    }
    
}
